// Name:
//	THM1176MatlabDriver.cpp
//
// Purpose:
//	Offer a way to communicate with a THM1176 using matlab.
//
// From Matlab:
//	
//	Command : THM1176MatlabDriver('list')	
//			  Returns the list of instrument found as Expanded Unaliased Name
//
//	Command : THM1176MatlabDriver('Open', <SerialNumber> | <'Expanded Unaliased Name'>)
//			  Returns a handle (> 0) or an error (-1) if the device cannot be open  
//
//  Command : THM1176MatlabDriver('Close', Handle)
//			  Closes the communication port
// 
//  Command : THM1176MatlabDriver('Read', Handle |, NumberOfMeasurementsRequested)
//			  Returns the measured magnitude of the magnetic field expressed in [Tesla]. 
//			  One measurement only is made. But, if NumberOfMeasurementsRequested is specified
//			  you will receive NumberOfMeasurementsRequested measurements.
//			  NumberOfMeasurementsRequested cannot exceed 2048.
//			  
//
// Change log:
//	24.04.14 JT	Original version 
//              Trigger mode is defaulted to IMMediate CONTinuous
//							The returned value is the Field Magnitude. 
//
//  04.11.16 JT Namespace THM1176MatlabDriver added to prevent char16_t to be defined twice.
//
//
//	To do:
//		Expand the model to handle the triggering mode, the units, the range and the various axis.
//
//  copyright:  
//    This code is distributed as-is and is free of charge. It is meant to be used by Metrolab Technology
//    customers for the sole purpose of interfacing instruments to Matlab.
//
//    This code uses the std library and no effort was made to optimize it.
//
//	  This code may be modified and distributed without restriction.
//
//    However, this license doesn't extend to the headers and libraries used by this code.  
//
//    Specifically :
//		visa.h is distributed under IVI Foundation Inc terms (http://www.ivifoundation.org/)
//      mex.h  is distributed under The Mathworks terms (http://www.mathworks.com/)
// 
//	  

// TABSTOPS = 4

#include <math.h>
#include <map>
#include <list>
#include <string>
#include <array>
#include <vector>
#include <iostream>
#include <algorithm>
#include <visa.h>

namespace THM1176MatlabDriver {
	#include "mex.h"


// Value-Defintions of the different commands handled
static enum eCommands { 
	kOpen = 1, // It is mandatory to start with 1, because any unmapped string will return an index value of 0.
	kClose,
	kRead,
	kList
};

// Map to associate the strings with the enum values
// Static variables values are kept from one call to another. 
// A matlab 'clear' command unload the dll from memory, resulting in the loss
// of these values.
static std::map<std::string, eCommands> _MapCommandsValues;

static bool								_Initialized = false;
static ViSession						_DefaultRessourcesManager;

// Map to associate the ViSession to the handle returned into the
// matlab workspace
typedef std::map<int,ViSession>			tRegisteredSessions;
static tRegisteredSessions				_RegisteredSessions;
static int								_RegisteredSessionsIndexer = 0;

void ListDevices(std::list<std::string> &rDeviceList, char* pSearchExpression)
{
	ViFindList				_RessourceList;
	ViChar					_RessourceDescriptor[VI_FIND_BUFLEN];
	ViUInt32				_NumberOfInstrumentsFound;
	ViUInt16				_InstrumentType;
	ViUInt16				_InterfaceNumber;
	ViChar					_RessourceClass[VI_FIND_BUFLEN];
	ViChar					_ExpandedUnaliasedName[VI_FIND_BUFLEN];
	ViChar					_AliasName[VI_FIND_BUFLEN];
	ViStatus				_Status;

	// Find all VXI instruments in the system
	_Status = viFindRsrc(_DefaultRessourcesManager, pSearchExpression, &_RessourceList, &_NumberOfInstrumentsFound, _RessourceDescriptor);
	
	if (_Status < VI_SUCCESS) {
		return;
	}
	
	// 
	// Collect the alias name into a list.
	//
	// This list, along with the maximum name size, will be used 
	// to create a string matrix to return the instrument list into 
	// the matlab workspace.
	//
	size_t _MaxStringLength = 0; 
	for (ViUInt32 _Index = 0; _Index < _NumberOfInstrumentsFound; _Index++)
	{	
		// Extract the information for the current ressource.
		// Be warned, the value is updated each time the method is called, we are not
		// given a pointer to a persistant value that viFindRsrc would have gathered
		// during its call. This means that we must actually must save the values of 
		// interest.
		_Status = viParseRsrcEx(_DefaultRessourcesManager, 
								_RessourceDescriptor, 
								&_InstrumentType, 									
								&_InterfaceNumber, 
								_RessourceClass, 
								_ExpandedUnaliasedName, 
								_AliasName);
		
		// Save the instrument name in a string
		std::string _AliasNameString(_ExpandedUnaliasedName);

		// Check if the size of this newly created string is greater than any others
		size_t _StringLength = _AliasNameString.length();
		_MaxStringLength = _MaxStringLength < _StringLength ? _StringLength : _MaxStringLength;  
		
		// Save the name for the future
		rDeviceList.push_back(_AliasNameString);
		
		// Get the next item in the list.
		_Status = viFindNext(_RessourceList, _RessourceDescriptor);
	}
	
}

void ListDevices(int			nlhs, 
				 mxArray*		plhs[],
                 int			nrhs, 
				 const mxArray*	prhs[])
{

  	if (nlhs > 1) 
      mexErrMsgTxt("Too many output arguments.");
	

	size_t _NumberOfInstrumentsFound = 0;

	// Gather the list of instruments present in the system.
	std::list<std::string>	_InstrumentList;
	ListDevices(_InstrumentList, "USB?*::0x1BFA::0x0498::?*::INSTR");

	// How many instruments were found ?
	_NumberOfInstrumentsFound = _InstrumentList.size();

	
	if (_NumberOfInstrumentsFound <= 0) {
		// No instrument present in the system, return an empty arry.
		plhs[0] = mxCreateCharArray(0, 0);
		return;
	}
	else
	{
		// 
		// From the alias name list, create a matrix of char into the matlab workspace
		//

		// Prepare an array of char pointer. It will be used as pointers to the heads of each 
		// instrument name string found during the enumeration.

  		const char **_pInstrumentNameArray = new const char*[_NumberOfInstrumentsFound];

		auto it = std::begin(_InstrumentList);
		for (int _i = 0; _i < (int) _NumberOfInstrumentsFound; _i++)
		{
			_pInstrumentNameArray[_i] = it->c_str();
			++it;
		}

		// Create the output matrix into the matlab workspace 
		plhs[0]= mxCreateCharMatrixFromStrings((mwSize)_NumberOfInstrumentsFound, (const char **)_pInstrumentNameArray);

		// Delete the array of char pointers.
		delete _pInstrumentNameArray;

	}
	_InstrumentList.clear();
}


void Open(	int				nlhs, 
			mxArray*		plhs[],
            int				nrhs, 
			const mxArray*	prhs[])
{
	int _HandleNumber = -1;

	// Check the number of input and output parameters
  	if (nlhs > 1) 
		mexErrMsgTxt("Too many output arguments.");

	if (nrhs != 2)
		mexErrMsgTxt("Incorrect number of input arguments.");

	std::list<std::string>	_InstrumentList;
	char * _pUnaliasedName = NULL;
	bool _MatlabAllocation = false;

	// If the input is a string, it means that we received the Expanded Unaliased Name 
	// of the device to open.
    if (mxIsChar(prhs[1]))
    {  
		// Get Expanded Unaliased Name
		_pUnaliasedName = mxArrayToString(prhs[1]);
		_MatlabAllocation = true;
	}
	// If the input is a number, it means that we received the serial number 
	// of the device to open.
	else if (mxIsNumeric(prhs[1]))
	{
		// We have no clue regarding the bus on which the instrument is located.
		// To get the Expanded Unaliased Name, we must collect all instruments

		// Get the serial number
		unsigned int _SerialNumber = static_cast<unsigned int> (mxGetScalar(prhs[1]));

		// Try to find the instrument.
		std::string				_SerialNumberString = std::to_string(_SerialNumber);
								_SerialNumberString.insert(_SerialNumberString.begin(),7-_SerialNumberString.length(),'0');
		std::string             _SearchExpression = "USB?*::0x1BFA::0x0498::" + _SerialNumberString + "::INSTR";

		ListDevices(_InstrumentList, (char*) _SearchExpression.c_str());

		if (_InstrumentList.size() == 1)
			_pUnaliasedName = (char*) _InstrumentList.begin()->c_str();
	}
	else
	{
		mexErrMsgTxt("Device to open must be specified as an 'Expanded Unaliased Name' or using its serial number.\n");
	}

	if (_pUnaliasedName != NULL)
	{
		// Try to open a session to the specified device.
		ViSession _Session;
		ViStatus  _Status;
		// Open the ressource
		_Status = viOpen(_DefaultRessourcesManager, (ViRsrc) _pUnaliasedName , NULL, NULL, &_Session);

		if (_Status < VI_SUCCESS)
		{
				_HandleNumber = -1;
		}
		else
		{
			// Register the opened session
			++_RegisteredSessionsIndexer;
			_RegisteredSessions[_RegisteredSessionsIndexer] = _Session;
			_HandleNumber = _RegisteredSessionsIndexer;
		}
	}

	// Free the string.
	if (_MatlabAllocation)
		mxFree(_pUnaliasedName);
	
	// Free the instrument list.
	_InstrumentList.clear();

	// Return the handle number
	// > 0 : Handle number
	// < 0 : Error
  	plhs[0] = mxCreateDoubleScalar(_HandleNumber);

}

bool Uncompress(char * _pGatheredData, double *_pSamples)
{
	unsigned int _Index = 0;

	// The compressed data must start with the hash sign
	if (_pGatheredData[_Index++] != '#')
		return false;

	// The second element is the size of the field containing the number
	// of data.
	unsigned int _FieldSize = _pGatheredData[_Index++]-'0';

	// The number of data contained in this chunk of data.
	std::string _NumberOfData;
	_NumberOfData.append(&_pGatheredData[_Index], _FieldSize);
	_Index+=_FieldSize;

	unsigned int _DataSize = std::stol(_NumberOfData) >> 2;
	for (unsigned int _i = 0; _i < _DataSize; _i++)
	{
		signed int _Data = 0;
		for (unsigned int _j = 0; _j <= 3; _j++)
		{
			unsigned char _Byte = _pGatheredData[_Index++];
			_Data = (_Data << 8) + _Byte;
		}
		_pSamples[_i] = _Data / 1e6;
	}
	return true;
}

void Read(	int				nlhs, 
			mxArray*		plhs[],
            int				nrhs, 
			const mxArray*	prhs[])
{
	// Check the number of input and output parameters
  	if (nlhs > 1) 
		mexErrMsgTxt("Too many output arguments.");

	if ((nrhs < 2) || (nrhs>3))
		mexErrMsgTxt("Incorrect number of input arguments.");

	int				_OperationStatus = -1;

	// If the user specified a number of sample to acquire, use it, otherwise, set
	// this value to 1.
	unsigned int	_NumberOfMeasurementsRequested = 1;
	if ((nrhs == 3) && (mxIsNumeric(prhs[2])))
		_NumberOfMeasurementsRequested = static_cast<unsigned int> (mxGetScalar(prhs[2]));

	// If the number of measurement requested doesn't exceed the maximum handled by the 
	// instrument, try to perform the operation.
	if (mxIsNumeric(prhs[1]) && (_NumberOfMeasurementsRequested <= 2048))
	{
		// Get the handle number and use it to retrieve the session.
		unsigned int _Handle = static_cast<unsigned int> (mxGetScalar(prhs[1]));
		ViSession _Session = _RegisteredSessions[_Handle];
		ViStatus  _Status = VI_SUCCESS;

		// Create the command requesting the number of measurement asked.
		std::string _Command = (":FORM INT;:MEAS:ARR? ") + std::to_string(_NumberOfMeasurementsRequested) +("\n");
		unsigned int _NumberOfBytesWritten = 0;

		// Send the command to the instrument
		_Status = viWrite(_Session, (ViBuf) _Command.c_str(), (ViUInt32) _Command.length(), (ViPUInt32) &_NumberOfBytesWritten);
		if (_Status >= VI_SUCCESS)
		{
			// We were able to send the request ... we must now collect what we sow.  
			
			// Create dynamically the buffers collecting the measurements;
			unsigned int _BufferSize = _NumberOfMeasurementsRequested*4+10;
			char*	_pGatheredData = new char [_NumberOfMeasurementsRequested*4+10];
					
			plhs[0] = mxCreateDoubleMatrix(1,_NumberOfMeasurementsRequested, mxREAL);

			unsigned int _NumberOfBytesReceived;

			_Status = viRead(_Session,(ViPBuf) _pGatheredData, _BufferSize, (ViPUInt32) &_NumberOfBytesReceived);
			if (_Status >= VI_SUCCESS)
				if (Uncompress(_pGatheredData,mxGetPr(plhs[0])))
					_OperationStatus = 0;

			delete _pGatheredData;
		}
	}
	else
	{
		// Return an empty matrix
		plhs[0] = mxCreateDoubleMatrix(0,0, mxREAL);
	}
}

void Close(	int				nlhs, 
			mxArray*		plhs[],
            int				nrhs, 
			const mxArray*	prhs[])
{
	// Check the number of input and output parameters
  	if (nlhs > 1) 
		mexErrMsgTxt("Too many output arguments.");

	if (nrhs != 2)
		mexErrMsgTxt("Incorrect number of input arguments.");

	int _OperationStatus = -1;

	if (mxIsNumeric(prhs[1]))
	{
		// Get the handle number
		unsigned int _Handle = static_cast<unsigned int> (mxGetScalar(prhs[1]));
		// Close the associated device
		if (viClose(_RegisteredSessions[_Handle]) >= VI_SUCCESS)
			_OperationStatus = 0;
	}
	
}

void Initialize()
{
	_MapCommandsValues["OPEN"]  = kOpen;
	_MapCommandsValues["CLOSE"] = kClose;
	_MapCommandsValues["READ"]  = kRead;
	_MapCommandsValues["LIST"]  = kList;
}

static void AtExit(void)
{
	// Clear the command container.
	_MapCommandsValues.clear();

	// Close all opened sesssions
	for (tRegisteredSessions::iterator _it=_RegisteredSessions.begin(); _it!=_RegisteredSessions.end(); ++_it)
		viClose(_it->second);

	// Clear the session container.
	_RegisteredSessions.clear();

	// Close the default ressource manager
	ViStatus  _Status = viClose(_DefaultRessourcesManager);
}

void mexFunction(int nlhs, 
				 mxArray *plhs[],
                 int nrhs, 
				 const mxArray *prhs[])
{
	if (!_Initialized)
	{
		// Register our exit function
		mexAtExit(AtExit);

		// Init the Commands map
		Initialize();

		// Try to open the default VISA ressource manager,
		// it is our entry point to all VISA operations.
		ViStatus  _Status = viOpenDefaultRM(&_DefaultRessourcesManager);
	
		if (_Status < VI_SUCCESS)
		    mexErrMsgTxt("Cannot open the default VISA ressource manager.\n");

		// Initialization is now terminated and does not need to
		// be performed again.
		_Initialized = true;
	}

    /* check for proper number of arguments */
    if (nrhs < 1) 
      mexErrMsgTxt("A command is required.");
    
    /* input must be a string */
    if ( mxIsChar(prhs[0]) != 1)
      mexErrMsgTxt("First Input must be a command string.");

	std::string _Command = mxArrayToString(prhs[0]);
	std::transform(_Command.begin(), _Command.end(), _Command.begin(), ::toupper);

	switch(_MapCommandsValues[_Command])
	{
		case kOpen:
			Open(nlhs,plhs,nrhs,prhs);
			break;

		case kClose:
			Close(nlhs,plhs,nrhs,prhs);
			break;
		
		case kRead:
			Read(nlhs,plhs,nrhs,prhs);
			break;
		
		case kList:
			ListDevices(nlhs,plhs,nrhs,prhs);
			break;

		default:
			printf("'%s' is an invalid string. s_mapStringValues now contains %d entries\n", _Command.c_str(), _MapCommandsValues.size());
			break;
	}
	return;
}

} // namespace THM1176MatlabDriver